from PyQt5.QtCore import QThread, pyqtSlot, QSize, pyqtSignal
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QLabel, QSizePolicy


class DummyGenerator(QThread):
    updateList = pyqtSignal(object)
    updateCameras = pyqtSignal(object)

    def __init__(self, parent=None):
        super(DummyGenerator, self).__init__(parent)
        self.num_cams = 7

    def getCameraList(self):
        return [x for x in range(0, self.num_cams)]

    def run(self):
        self.updateList.emit(self.getCameraList())


class DummyCamera(QLabel):

    def __init__(self, source=None, size=300, name="default", scaled=False, parent=None):
        super(DummyCamera, self).__init__(parent)
        self.name = name
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setMinimumSize(QSize(50, 50))
        self.setImage()

    @pyqtSlot(object, name="size")
    def setFrameSize(self, size):
        pass
        # self.setPixmap(self.px.scaled(size, size))
        self.setPixmap(self.px.scaled(size, size))  # , Qt.KeepAspectRatio))

    def setImage(self):
        # r = requests.get("https://picsum.photos/500").content
        # qp = QPixmap()
        # qp.loadFromData(r)

        qp = QPixmap("resources/ent.jpg")
        self.px = qp
        qp = qp.scaled(300, 300)  # , Qt.KeepAspectRatio)
        self.setPixmap(qp)



import shutil
import sys
from os.path import exists

import click
import math
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from click_default_group import DefaultGroup

from piveilance import util
from piveilance._version import __version__
from piveilance.cameras.camera import *
from piveilance.cameras.dummy import *
from piveilance.config import *
from piveilance.resources import get_resource


@click.group(cls=DefaultGroup, default='start', default_if_no_args=True)
def cli():
    pass


@cli.command()
@click.pass_context
@click.option('-c', '--config',
              help="point to config ini by name",
              type=click.Path(exists=True),
              default=None, )
@click.option('-f', '--fullscreen',
              help="toggle fullscreen",
              is_flag=True,
              default=None)
@click.option('-s', '--stretch',
              help="toggle stretch",
              is_flag=True,
              default=None)
def start(ctx, **kwargs):
    app = QApplication(sys.argv)
    _ = PiWndow(ctx, **kwargs)
    sys.exit(app.exec_())


class PiWndow(QMainWindow):
    setCamSize = pyqtSignal(object)
    setCamOptions = pyqtSignal(object)
    resized = pyqtSignal()

    def __init__(self, ctx, **cli_options):
        super().__init__()

        title = "Pi Veilance"
        stylesheet = get_resource("styles.qss")
        icon = get_resource("icon.ico")

        cliArgs, configFile = self.parseCLIArgs(cli_options)
        self.config = Config.from_yaml(configFile).merge(cliArgs)
        self.camConfig = self.config.get_config('cameras')
        self.viewConfig = self.config.get_config('view')

        self.setCamClass()
        self.beginDataFlows()
        self.initWindow(stylesheet, title, icon)
        self.showFullScreen() if self.fullscreen else self.show()

    def parseCLIArgs(self, cli_options):

        configFile = cli_options.get('config') or "config.yaml"
        if not exists(configFile):
            shutil.copy(get_resource(configFile), ".")

        cli = {}
        for o in cli_options:
            v = cli_options[o]
            if not v:
                continue
            elif v and o in ['fullscreen']:
                Config.merge_dict(cli, Config.make_dict(['view', o], v))
            elif v and o in ['stretch']:
                Config.merge_dict(cli, Config.make_dict(['cameras', o], v))

        return cli, configFile

    def setCamClass(self):
        self.camClass = self.camConfig.get('type', 'picam')
        if self.camClass == "picam":
            self.globalGen = PiCamGenerator
        elif self.camClass == "dummy":
            self.globalGen = DummyGenerator
        else:
            raise TypeError("Unknown camera class: " + self.camClass)

    def initWindow(self, stylesheet, title, icon):
        self.fullscreen = self.viewConfig.get_bool('fullscreen', False)
        self.widget = QWidget()
        self.resize(1024, 768)
        qr = self.frameGeometry()
        cp = QDesktopWidget().availableGeometry().center()
        qr.moveCenter(cp)
        self.setWindowTitle(title)
        self.setWindowIcon(QIcon(icon))
        self.setCentralWidget(self.widget)
        self.statusBar().showMessage('Version: ' + __version__)
        self.grid = QGridLayout()
        self.grid.setSpacing(0)
        self.grid.setContentsMargins(0, 0, 0, 0)
        self.widget.setLayout(self.grid)
        self.setStyleSheet(open(stylesheet, "r").read())
        self.resized.connect(self.setCameraGrid)

    def resizeEvent(self, event):
        """
        Overridden method - resize the gride on window resize
        :param event:
        :return:
        """
        self.resized.emit()
        return super(PiWndow, self).resizeEvent(event)

    def contextMenuEvent(self, event):
        """
        Override to add menu quit option
        :param event:
        :return:
        """

        cmenu = QMenu(self)
        quitAct = cmenu.addAction("Quit")
        fullScreenAct = cmenu.addAction("Toggle fullscreen")
        stretchAct = cmenu.addAction("Toggle stretch")

        cropMenu = cmenu.addMenu("Crop ratio")
        entries = [str(float(i / 10)) for i in range(0, 11)]

        for e in entries:
            a = cropMenu.addAction(e)
            a.name = "crop_" + e

        action = cmenu.exec_(self.mapToGlobal(event.pos()))
        if action == quitAct:
            qApp.quit()
        elif action == fullScreenAct:
            if self.isFullScreen():
                self.showNormal()
            else:
                self.showFullScreen()
        elif action == stretchAct:
            current = self.camConfig.get_bool('stretch', False)
            self.camConfig['stretch'] = not current
            self.setCameraGrid()

        elif hasattr(action, 'name') and action.name.startswith("crop"):
            r = float(action.name.split("_")[1])
            self.camConfig['crop_ratio'] = r
            self.setCameraGrid()

    def beginDataFlows(self):
        self.camlist = []
        self.cols = 0
        self.max_cameras = self.camConfig.get_int('max_allowed', 0)
        self.generator = self.globalGen(self.camConfig)
        self.generator.updateList.connect(self.setCameraGrid)
        self.generator.start()

    @pyqtSlot(object, name="camgrid")
    @pyqtSlot(name="resize")
    def setCameraGrid(self, camlist=None, adjust=False):

        if camlist and not util.areSetsEqual(camlist, self.camlist):
            adjust = True

        # Must keep as instance var in case called with none
        self.camlist = camlist or self.camlist

        if not self.camlist:
            return

        if self.max_cameras > 0:
            self.camlist = self.camlist[0:self.max_cameras]

        # see if there are new column count as calculated by compute
        width = self.widget.frameGeometry().width()
        height = self.widget.frameGeometry().height()
        new_cols, dimensions, self.camConfig['size'] = \
            util.computeGrid(len(self.camlist), width, height)

        if not self.camConfig.get_bool('stretch'):
            self.grid.setContentsMargins(*dimensions)
        else:
            self.grid.setContentsMargins(0, 0, 0, 0)

        # Cameras must be repositioned
        if adjust or new_cols != self.cols:

            # calculate new rows and positions
            self.cols = new_cols
            rows = math.ceil(len(self.camlist) / new_cols)
            positions = [(i, j) for i in range(rows) for j in range(new_cols)]

            # clear the layout
            for i in reversed(range(self.grid.count())):
                self.grid.takeAt(i).widget().deleteLater()

            # add the cameras
            for i, name in enumerate(self.camlist):
                cam = self.generator.createCamera(name, self.camConfig)
                self.setCamSize.connect(cam.setFrameSize)
                self.setCamOptions.connect(cam.setOptions)
                self.grid.addWidget(cam, *positions[i])
        self.setCamSize.emit(self.camConfig['size'])
        self.setCamOptions.emit(self.camConfig)


if __name__ == '__main__':
    cli()



import ast
import json
import os
import re
from collections import UserDict, Mapping
from copy import deepcopy
from datetime import datetime, timedelta
from io import StringIO
from json import JSONDecodeError

import yaml
from dateutil import parser
from yaml.scanner import ScannerError


class Config(UserDict):

    def __init__(self, data):
        super(Config, self).__init__(data)

    def get_config(self, name):
        return Config(self.get(name, {}))

    def get_as(self, name, default=None, required=False, as_type=None):
        val = self.get(name, default)
        if required and not val:
            raise AssertionError("Value for " + name + " required but not given")
        return Parser.parse_type(val, as_type) if as_type else val

    def get_bool(self, name, default=False, required=False):
        val = self.get_as(name, default, required)
        return str(val).lower() in ['true', '1']

    def get_string(self, name, default=False, required=False, decode=False):
        val = self.get_as(name, default, required, str)
        return Parser.decode(val, True) if decode else val

    def get_int(self, name, default=False, required=False):
        return self.get_as(name, default, required, int)

    def get_float(self, name, default=False, required=False):
        return self.get_as(name, default, required, float)

    def get_list(self, name, default=False, required=False):
        return self.get_as(name, default, required, list)

    def get_dict(self, name, default=False, required=False):
        return self.get_as(name, default, required, dict)

    def merge(self, update):
        self.merge_dict(self.data, update)
        return self

    def get_path(self, name, default=None, required=False):
        raw = self.get_as(name, default, required)
        return os.sep.join(re.split('[/\\\\]+', raw)) if raw else None

    def load_data_for_keys(self, parent, dir=''):
        for r, path in parent.items():
            if path.split('.')[-1].lower() in ['yml', 'yaml']:
                parent[r] = self.load_yaml(os.path.join(dir, path))
            else:
                parent[r] = os.path.join(dir, path)

    @classmethod
    def from_yaml(cls, props_file):
        return Config(cls.load_yaml(props_file))

    @staticmethod
    def load_yaml(yaml_file):
        with open(yaml_file, 'r') as file:
            return yaml.safe_load(file)

    @staticmethod
    def make_dict(keylist, value):
        tree_dict = {}
        for i, key in enumerate(reversed(keylist)):
            val = value if i == 0 else tree_dict
            tree_dict = {
                key: val}
        return tree_dict

    @staticmethod
    def merge_dict(d1, d2, immutable=False):
        """
        # Combine dictionaries recursively
        # preserving the originals
        # assumes d1 and d2 dictionaries!!
        :param d1: original dictionary
        :param d2: update dictionary
        :return:
        """

        d1 = {} if d1 is None else d1
        d2 = {} if d2 is None else d2
        d1 = deepcopy(d1) if immutable else d1

        for k in d2:
            # if d1 and d2 have dict for k, then recurse
            # else assign the new value to d1[k]
            if (k in d1 and isinstance(d1[k], Mapping)
                    and isinstance(d2[k], Mapping)):
                Config.merge_dict(d1[k], d2[k])
            else:
                d1[k] = d2[k]
        return d1


class Parser():

    @staticmethod
    def parse_type(value, as_type):

        try:
            if as_type == bool:
                return Parser.parse_bool(value)
            elif as_type == list:
                return Parser.parse_list(value)
            elif as_type == dict:
                return Parser.parse_dict(value)
            elif as_type == timedelta:
                return Parser.parse_time_delta(value)
            elif as_type == datetime:
                return Parser.parse_datetime(value)
            elif as_type == "json":
                return Parser.parse_yaml_json(value, data_type="json")
            elif as_type == "yaml":
                return Parser.parse_yaml_json(value, data_type="yaml")
            else:
                return as_type(value)
        except:
            raise TypeError("Error parsing {0} as type {1}".format(value, as_type))

    @staticmethod
    def parse_datetime(date):
        if isinstance(date, datetime):
            return date
        if not isinstance(date, str):
            return TypeError("Cannot parse date '{0}' from type: {1}".
                             format(str(datetime), str(type(datetime))))
        return parser.parse(date)

    @staticmethod
    def parse_bool(value):
        if not value:
            return False
        if isinstance(value, bool):
            return value
        if value in ["1", "0", 1, 0]:
            return str(value) == "1"
        elif value.lower() in ["true", "false"]:
            return value == "true"
        raise TypeError("Cannot cast: " + str(value) + " to boolean")

    @staticmethod
    def parse_list(value):
        if isinstance(value, set):
            return list(value)
        else:
            return Parser.parse_collection(value, list)

    @staticmethod
    def parse_dict(value):
        return Parser.parse_collection(value, dict)

    @staticmethod
    def parse_collection(value, as_type):
        if value is None:
            return as_type()
        elif isinstance(value, as_type):
            return value
        else:
            try:
                value = value.decode()
            except:
                pass
            try:
                return json.loads(value)
            except:
                pass
            try:
                return ast.literal_eval(value)
            except:
                pass
        return as_type(value)

    @staticmethod
    def parse_time_delta(value):
        if isinstance(value, timedelta):
            return value
        if value is not None:
            value = str(value)
            intervals = re.split('[(days, ):]+', value)
            if len(intervals) >= 3:
                d = float(intervals[0]) if len(intervals) == 4 else 0.0
                s = float(intervals[-1])
                m = float(intervals[-2])
                h = float(intervals[-3])
                value = timedelta(hours=h, minutes=m, seconds=s, days=d)
        return value

    @staticmethod
    def parse_yaml_json(data, data_type="json"):
        if not data:
            return {}
        try:
            if data_type == 'json':
                return json.loads(data)
            elif data_type in ['yml', 'yaml']:
                return yaml.load(StringIO(data))
        except (JSONDecodeError, ScannerError, ValueError, AttributeError) as e:
            raise ConfigParsingException("Error parsing config from data: ", e)
        raise ConfigParsingException("Format not allowed: " + data_type)

    @staticmethod
    def correct_path(path):
        path = re.split('[/\\\\]+', path)
        return os.sep.join(path)

    @staticmethod
    def decode(text, lower=False):
        if not text:
            return
        try:
            text = text.decode()
        except:
            pass

        text = str(text).strip()
        return text.lower() if lower else text

    @staticmethod
    def compare_str(a, b):
        return Parser.decode(a, True) == Parser.decode(b, True)



class ConfigParsingException(Exception):
    __name__ = "ConfigParsingException"

    def __init__(self, message, error=None):
        Exception.__init__(self)
        self.args = (message,)
        self.message = message
        self.error = error



from piveilance.config import Config
import yaml

def test_config():
    dict_a = {
        'a': 'b',
        'c': ['d', 'e'],
        'd': {'x': 't'},
        'e': {
            'a': [1, 2, '3'],
            'b1': {
                'a1': 'c1',
                'a2': ('a3', 'a4')
            }
        },
        'f': None
    }

    dict_b = {
        'x': {'y': 'z'},
        'a': 'bb',
        'c': ['dd', 'e'],
        'd': {'j': 'u'},
        'e': {

            'b1': {
                'a2': ('a3', 'a4')
            }
        },
        'f': {'g'}
    }

    r1 =  Config.merge_dict(dict_a, dict_b)

    Config.merge_dict(dict_a, dict_b, immutable=False)

    assert dict_a == r1 == {
        'a': 'bb',
        'c': ['dd', 'e'],
        'd': {
            'x': 't',
            'j': 'u'
        },
        'e': {
            'a': [1, 2, '3'],
            'b1': {
                'a1': 'c1',
                'a2': ('a3', 'a4')
            }
        },
        'f': {'g'},
        'x': {'y': 'z'}
    }

def test_complex():

    with open("complex_dict.yml") as config:
        complex = yaml.safe_load(config)

    with open("merge_config.yml") as config:
        merge = yaml.safe_load(config)

    with open('result.yml') as output:
        actual = yaml.safe_load(output)

    combined = Config.merge_dict(complex, merge)
    assert combined == actual

    # with open('temp.yml', 'w') as output:
    #     yaml.dump(combined, output)
    # print()





import base64
import json
import time
from copy import deepcopy

import requests
from PyQt5.QtCore import QThread, pyqtSlot, QSize, pyqtSignal
from PyQt5.QtGui import QPixmap, QImage
from PyQt5.QtWidgets import QLabel, QSizePolicy

from piveilance.config import Parser
from piveilance.util import ImageManip


class PiCamGenerator(QThread):
    updateList = pyqtSignal(object)
    updateCameras = pyqtSignal(object)

    def __init__(self, config, parent=None):
        super(PiCamGenerator, self).__init__(parent)
        self.list_url = config.get('list_url')
        self.data_url = config.get('data_url')
        self.sleep = config.get_float('update_interval', 0.1)
        self.list_refresh = config.get_float('list_refresh', 10)

    def getCameraList(self):
        cams = requests.get(self.list_url).content
        return json.loads(cams)

    def update(self):
        img = requests.get(self.data_url)
        data = json.loads(img.content)
        camData = {v['source']: v for v in data}
        self.updateCameras.emit(camData)

    def run(self):
        start = time.time()
        self.updateList.emit(self.getCameraList())
        while True:
            self.update()
            time.sleep(self.sleep)
            if time.time() - start > self.list_refresh:
                self.updateList.emit(self.getCameraList())
                start = time.time()

    def createCamera(self, name, config):
        cam = Camera(name, config)
        self.updateCameras.connect(cam.setImage)
        return cam


class Camera(QLabel):
    def __init__(self,
                 name="default",
                 options=None,
                 parent=None):

        super(Camera, self).__init__(parent)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setMinimumSize(QSize(50, 50))
        self.px = None
        self.name = name
        self.setOptions(options)

    @pyqtSlot(object, name="reconfigure")
    def setOptions(self, options):
        self.options = deepcopy(options)
        for o, v in self.options.get_dict('overrides').items():
            if Parser.compare_str(o, self.name):
                self.options.update(v)

        self.size = self.options.get_int('size')
        self.crop_ratio = self.options.get_float('crop_ratio')
        self.direction = self.options.get_string('direction', 'right', decode=True)
        self.setScaledContents(self.options.get_bool('stretch'))

        if self.crop_ratio < 0 or self.crop_ratio > 1:
            raise ValueError("Crop cannot be negative or inverse (>1)")

    @pyqtSlot(object, name="size")
    def setFrameSize(self, size):
        if self.px:
            self.setPixmap(self.px.scaled(size, size))
            self.size = size

    @pyqtSlot(object, name="setimage")
    def setImage(self, camData=None):
        if self.name in camData:
            data = self.getImage(camData[self.name]['image'])

            img = QImage()
            img.loadFromData(data)

            if img.width() > img.height() and self.crop_ratio != 0:
                crop = (img.width() - img.height()) * self.crop_ratio
                img = ImageManip.crop_direction(img, crop, self.direction)


            self.px = QPixmap().fromImage(img)
            self.setFrameSize(self.size)

    def getImage(self, data):
        byte = data.encode('utf-8')
        return base64.decodebytes(byte)



import math


def computeGrid(camCount, width, height):
    cols, rows, sideLength = calculateCols(camCount, width, height)
    remainder_height = height - sideLength * rows

    if remainder_height < 0:
        sideLength = sideLength - abs(remainder_height / rows)

    rheight = max(height - sideLength * rows, 0) / 2
    vheight = max(width - sideLength * cols, 0) / 2
    dimensions = (vheight, rheight, vheight, rheight)
    return cols, dimensions, sideLength


def calculateCols(camCount, width, height):
    """
    # Iterative computation to determine actual cols

    """
    # Start by initial guess that ncols = ncams
    cols = rows = sideLength = camCount
    while (cols > 0):
        rows = math.ceil(camCount / cols)
        sideLength = width / cols
        if ((height - sideLength * rows) < sideLength
                and (rows * cols - camCount) <= 1):
            break
        cols -= 1
    cols = 1 if cols < 1 else cols
    return cols, rows, sideLength


def areSetsEqual(a, b):
    a = a or set()
    b = b or set()
    a = set(a)
    b = set(b)
    return (a - b) == (b - a)


class ImageManip():

    @classmethod
    def crop_direction(cls, image, amount, direction='center'):
        if direction == 'right':
            return ImageManip.crop(image, 0, 0, 0, amount)
        elif direction == 'hcenter':
            return ImageManip.crop(image, 0, amount / 2, 0, amount / 2)
        elif direction == 'vcenter':
            return ImageManip.crop(image, amount / 2, 0, amount / 2, 0)
        elif direction == 'left':
            return ImageManip.crop(image, 0, 0, 0, amount)
        elif direction == 'top':
            return ImageManip.crop(image, amount, 0, 0, 0)
        elif direction == 'bottom':
            return ImageManip.crop(image, 0, 0, amount, 0)
        elif direction == 'square':
            return ImageManip.crop(image, amount/4, amount/4, amount/4, amount/4)
        raise ValueError("Invalid direction: " + direction)

    @classmethod
    def crop(cls, image, top, left, bottom, right):
        w = image.width() - left - right
        h = image.height() - top - bottom
        return image.copy(left, top, w, h)

    @classmethod
    def cropCenter(cls, image, size):
        return cls.crop(image, size, size, size, size)



from setuptools import setup, find_packages
from piveilance._version import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup_deps = [
          'pyinstaller',
          'wheel',
          'twine'
      ],

test_deps = ['pytest']

setup(name='piveilance',
      version=__version__,
      description='PiCam Surveilance App',
      long_description=long_description,
      long_description_content_type="text/markdown",
      classifiers=[],
      url='https://github.com/vossenv/piveilance',
      maintainer='Danimae Vossen',
      maintainer_email='vossen.dm@gmail.com',
      license='MIT',
      packages=find_packages(),
      package_data={
          'piveilance': ['resources/*'],
      },
      install_requires=[
          'click',
          'requests',
          'python_dateutil',
          'click-default-group',
      ],
      extras_require={
          ':sys_platform=="win32"': [
              'pyqt5',
              'pyqt5-sip'
          ],
          'setup': setup_deps,
      },
      setup_requires=setup_deps,
      tests_require=test_deps,
      entry_points={
          'console_scripts': [
              'piveilance = piveilance.app:cli',
          ]
      },
      )
